//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleLayoutItem.cpp
//! @brief     Implements class ParticleLayoutItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Types/DoubleDescriptor.h"

ParticleLayoutItem::ParticleLayoutItem() : SessionItem(M_TYPE)
{
    setToolTip("A layout of particles");

    addProperty(P_TOTAL_DENSITY, 0.01);
    getItem(P_TOTAL_DENSITY)->setDecimals(10);
    addProperty(P_WEIGHT, 1.0);

    registerTag(T_PARTICLES, 0, -1,
                {ParticleItem::M_TYPE, ParticleCoreShellItem::M_TYPE,
                 ParticleCompositionItem::M_TYPE, MesoCrystalItem::M_TYPE});
    setDefaultTag(T_PARTICLES);
    registerTag(T_INTERFERENCE, 0, 1,
                {Interference1DLatticeItem::M_TYPE, Interference2DLatticeItem::M_TYPE,
                 Interference2DParaCrystalItem::M_TYPE, InterferenceFinite2DLatticeItem::M_TYPE,
                 InterferenceHardDiskItem::M_TYPE, InterferenceRadialParaCrystalItem::M_TYPE});
}

DoubleDescriptor ParticleLayoutItem::ownDensity() const
{
    DoubleDescriptor d(getItem(P_TOTAL_DENSITY), Unit::nanometerPowerMinus2);
    d.tooltip = "Number of particles per area (particle surface density).\n "
                "Should be defined for disordered and 1d-ordered particle collections.";
    return d;
}

double ParticleLayoutItem::totalDensity() const
{
    if (!totalDensityIsDefinedByInterference())
        return ownDensity();

    const auto* interferenceItem = getItem(T_INTERFERENCE);
    ASSERT(interferenceItem);

    if (const auto* interLatticeItem =
            dynamic_cast<const Interference2DAbstractLatticeItem*>(interferenceItem)) {
        Lattice2DItem* latticeItem = interLatticeItem->latticeType().currentItem();
        try {
            const double area = latticeItem->unitCellArea();
            return area == 0.0 ? 0.0 : 1.0 / area;
        } catch (const std::exception&) {
            // nothing to do here; new exception will be caught during job execution
            return 0.0;
        }
    } else if (const auto* interHDItem =
                   dynamic_cast<const InterferenceHardDiskItem*>(interferenceItem))
        return interHDItem->density();
    else {
        ASSERT(false);
        return 0.0;
    }
}

DoubleDescriptor ParticleLayoutItem::weight() const
{
    DoubleDescriptor d(getItem(P_WEIGHT), Unit::unitless);
    d.tooltip = "Weight of this particle layout.\n"
                "Should be used when multiple layouts define different domains in the sample.";
    return d;
}

QVector<ItemWithParticles*> ParticleLayoutItem::particles() const
{
    return items<ItemWithParticles>(T_PARTICLES);
}

void ParticleLayoutItem::addParticle(ItemWithParticles* particle)
{
    model()->moveItem(particle, this, -1, T_PARTICLES);
}

void ParticleLayoutItem::removeParticle(ItemWithParticles* particle)
{
    model()->removeItem(particle);
}

QVector<ItemWithParticles*> ParticleLayoutItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    for (auto* p : particles())
        result << p << p->containedItemsWithParticles();
    return result;
}

SelectionDescriptor<InterferenceItem*> ParticleLayoutItem::interference() const
{
    SelectionDescriptor<InterferenceItem*> d;

    // we need a special filling for this selection descriptor (not just from a GroupItem), since
    // the interference is stored in a child, which can be present or not.

    static QVector<QPair<QString, QString>> map;

    // initialize if not done so far
    if (map.isEmpty()) {
        map << qMakePair(QString("None"), QString(""));
        for (const auto& modelType : ItemCatalog::interferenceFunctionTypes())
            map << qMakePair(ItemCatalog::instance().uiInfo(modelType).menuEntry, modelType);
    }

    d.label = "Interference function";

    for (auto [title, type] : map)
        d.options << title;

    d.currentItem = [=] { return dynamic_cast<InterferenceItem*>(getItem(T_INTERFERENCE)); };

    d.currentIndexSetter = [=](int current) {
        if (auto* item = getItem(T_INTERFERENCE))
            model()->removeItem(item);
        if (current > 0)
            model()->insertNewItem(map[current].second, const_cast<ParticleLayoutItem*>(this), -1,
                                   T_INTERFERENCE);
    };

    d.currentIndexGetter = [=]() {
        if (auto* item = dynamic_cast<InterferenceItem*>(getItem(T_INTERFERENCE)))
            for (int i = 1; i < map.size(); i++)
                if (map[i].second == item->modelType())
                    return i;

        return 0;
    };

    return d;
}

void ParticleLayoutItem::setInterference(InterferenceItem* interference)
{
    model()->moveItem(interference, this, -1, T_INTERFERENCE);
}

void ParticleLayoutItem::removeInterference()
{
    if (auto* item = getItem(T_INTERFERENCE))
        model()->removeItem(item);
}

void ParticleLayoutItem::enableDensity(bool b)
{
    getItem(P_TOTAL_DENSITY)->setEnabled(b);
}

bool ParticleLayoutItem::totalDensityIsDefinedByInterference() const
{
    const auto* interferenceItem = getItem(T_INTERFERENCE);
    return dynamic_cast<const Interference2DAbstractLatticeItem*>(interferenceItem)
           || dynamic_cast<const InterferenceHardDiskItem*>(interferenceItem);
}
