//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/FormFactorComposition.cpp
//! @brief     Implements class FormFactorComposition.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/FormFactorComposition.h"
#include "Base/Util/Algorithms.h"

FormFactorComposition::~FormFactorComposition()
{
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        delete m_formfactors[index];
}

FormFactorComposition* FormFactorComposition::clone() const
{
    auto* result = new FormFactorComposition();
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        result->addFormFactor(*m_formfactors[index]);
    return result;
}

double FormFactorComposition::radialExtension() const
{
    double result{0.0};
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        result += m_formfactors[index]->radialExtension();
    return result;
}

double FormFactorComposition::bottomZ(const IRotation* rotation) const
{
    if (m_formfactors.empty())
        throw std::runtime_error("FormFactorComposition::bottomZ() -> Error: "
                                 "'this' contains no form factors.");
    return BaseUtils::algo::min_value(
        m_formfactors.begin(), m_formfactors.end(),
        [&rotation](IFormFactor* ff) { return ff->bottomZ(rotation); });
}

double FormFactorComposition::topZ(const IRotation* rotation) const
{
    if (m_formfactors.empty())
        throw std::runtime_error("FormFactorComposition::topZ() -> Error: "
                                 "'this' contains no form factors.");
    return BaseUtils::algo::max_value(m_formfactors.begin(), m_formfactors.end(),
                                      [&rotation](IFormFactor* ff) { return ff->topZ(rotation); });
}

void FormFactorComposition::addFormFactor(const IFormFactor& formfactor)
{
    m_formfactors.push_back(formfactor.clone());
}

void FormFactorComposition::setAmbientMaterial(const Material& material)
{
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        m_formfactors[index]->setAmbientMaterial(material);
}

complex_t FormFactorComposition::theFF(const WavevectorInfo& wavevectors) const
{
    complex_t result(0.0, 0.0);
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        result += m_formfactors[index]->theFF(wavevectors);
    return result;
}

Eigen::Matrix2cd FormFactorComposition::thePolFF(const WavevectorInfo& wavevectors) const
{
    Eigen::Matrix2cd result = Eigen::Matrix2cd::Zero();
    for (size_t index = 0; index < m_formfactors.size(); ++index)
        result += m_formfactors[index]->thePolFF(wavevectors);
    return result;
}
