//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/MesoCrystal.cpp
//! @brief     Implements class MesoCrystal.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/MesoCrystal.h"
#include "Base/Util/Assert.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/FormFactorCrystal.h"
#include "Sample/Particle/ParticleInSlice.h"
#include "Sample/Scattering/IBornFF.h"
#include "Sample/Scattering/Rotations.h"

MesoCrystal::MesoCrystal(Crystal* crystal, IBornFF* formfactor)
    : m_crystal(crystal), m_meso_formfactor(formfactor)
{
}

MesoCrystal::MesoCrystal(const Crystal& crystal, const IBornFF& formfactor)
    : MesoCrystal(crystal.clone(), formfactor.clone())
{
}

MesoCrystal::~MesoCrystal() = default;

MesoCrystal* MesoCrystal::clone() const
{
    auto* result = new MesoCrystal(m_crystal->clone(), m_meso_formfactor->clone());
    result->setAbundance(m_abundance);
    if (m_rotation)
        result->setRotation(*m_rotation);
    result->setParticlePosition(m_position);
    return result;
}

std::vector<const INode*> MesoCrystal::nodeChildren() const
{
    return std::vector<const INode*>()
           << IParticle::nodeChildren() << m_crystal << m_meso_formfactor;
}

ParticleInSlice MesoCrystal::createParticleInSlice(const ZLimits& limits) const
{
    ASSERT(m_crystal && m_meso_formfactor);

    double unit_cell_volume = m_crystal->lattice()->unitCellVolume();
    if (unit_cell_volume <= 0)
        return {};

    ParticleInSlice result;

    std::unique_ptr<IDecoratableBorn> tem_ff(
        m_meso_formfactor->createSlicedFormFactor(limits, m_rotation.get(), m_position));

    const std::unique_ptr<Crystal> new_crystal(
        m_crystal->transformed(m_rotation.get(), m_position));
    const std::unique_ptr<IFormFactor> new_basis_ff(new_crystal->basis()->createFormFactor());

    result.sliced_ff.reset(new FormFactorCrystal(*new_crystal->lattice(), *new_basis_ff, *tem_ff,
                                                 new_crystal->position_variance()));

    Admixtures& regions = result.admixtures;
    ASSERT(regions.empty());
    for (const auto& particle : m_crystal->basis()->decompose()) {
        ParticleInSlice pis = particle->createParticleInSlice({});
        regions.insert(regions.end(), pis.admixtures.begin(), pis.admixtures.end());
    }
    for (auto& region : regions)
        region.fraction *= m_meso_formfactor->volume() / unit_cell_volume;

    return result;
}

const Crystal& MesoCrystal::particleStructure() const
{
    ASSERT(m_crystal);
    return *m_crystal;
}
