//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Scattering/DecoratedFF.cpp
//! @brief     Implements interface class DecoratedFF.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Scattering/DecoratedFF.h"
#include "Base/Vector/RotMatrix.h"
#include "Sample/Material/Material.h"
#include "Sample/Material/MaterialFactoryFuncs.h"
#include "Sample/Scattering/Rotations.h"

DecoratedFF::DecoratedFF(IDecoratableBorn* ff, const Material& material,
                         const Material& ambient_material, const R3* position,
                         const RotMatrix* rotMatrix)
    : m_ff(std::move(ff))
    , m_material(std::make_unique<const Material>(material))
    , m_ambient_material(std::make_unique<Material>(ambient_material))
    , m_position(std::move(position))
    , m_rotMatrix(std::move(rotMatrix))
{
}

DecoratedFF::DecoratedFF(const IDecoratableBorn& ff)
    : DecoratedFF(ff.clone(), HomogeneousMaterial(), HomogeneousMaterial(), nullptr, nullptr)
{
}

DecoratedFF::~DecoratedFF() = default;

DecoratedFF* DecoratedFF::clone() const
{
    return new DecoratedFF(m_ff->clone(), *m_material,
                           *m_ambient_material,
                           m_position ? new R3(*m_position) : nullptr,
                           m_rotMatrix ? new RotMatrix(*m_rotMatrix) : nullptr);
}

void DecoratedFF::setMaterial(const Material& material)
{
    m_material = std::make_unique<Material>(material);
}

void DecoratedFF::setAmbientMaterial(const Material& ambient_material)
{
    m_ambient_material = std::make_unique<Material>(ambient_material);
}

void DecoratedFF::setPosition(const R3& position)
{
    m_position = std::make_unique<R3>(position);
}

void DecoratedFF::setRotMatrix(const RotMatrix& rotMatrix)
{
    m_rotMatrix = std::make_unique<RotMatrix>(rotMatrix);
}

complex_t DecoratedFF::theFF(const WavevectorInfo& wavevectors) const
{
    complex_t result = m_ff->theFF(wavevectors);
    if (m_material && m_ambient_material)
        result = (m_material->scalarSubtrSLD(wavevectors)
                  - m_ambient_material->scalarSubtrSLD(wavevectors)) * result;
    return result;
}

Eigen::Matrix2cd DecoratedFF::thePolFF(const WavevectorInfo& wavevectors) const
{
    Eigen::Matrix2cd result = m_ff->thePolFF(wavevectors);
    if (m_material && m_ambient_material) {
        // the conjugated linear part of time reversal operator T
        // (T=UK with K complex conjugate operator and U is linear)
        Eigen::Matrix2cd time_reverse_conj = Eigen::Matrix2cd::Zero();
        time_reverse_conj(0, 1) = 1.0;
        time_reverse_conj(1, 0) = -1.0;
        // the interaction and time reversal taken together:
        Eigen::Matrix2cd V_eff = time_reverse_conj
            * (m_material->polarizedSubtrSLD(wavevectors)
               - m_ambient_material->polarizedSubtrSLD(wavevectors));
        result *= V_eff;
    }
    return result;
}

/*

complex_t DecoratedFF::theFF(const WavevectorInfo& arg) const
{
    WavevectorInfo wavevectors = arg;
    if (m_rotMatrix)
        wavevectors = wavevectors.transformed(m_rotMatrix->getInverse());
    complex_t result = m_ff->theFF(wavevectors);
    if (m_material && m_ambient_material)
        result *= m_material->scalarSubtrSLD(wavevectors)
                  - m_ambient_material->scalarSubtrSLD(wavevectors);
    if (m_position)
        return result *= exp_I(m_position->dot(wavevectors.getQ()));
    return result;
}

Eigen::Matrix2cd DecoratedFF::thePolFF(const WavevectorInfo& arg) const
{
    WavevectorInfo wavevectors = arg;
    if (m_rotMatrix)
        wavevectors = wavevectors.transformed(m_rotMatrix->getInverse());
    Eigen::Matrix2cd result = m_ff->thePolFF(wavevectors);
    if (m_material && m_ambient_material) {
        // the conjugated linear part of time reversal operator T
        // (T=UK with K complex conjugate operator and U is linear)
        Eigen::Matrix2cd time_reverse_conj = Eigen::Matrix2cd::Zero();
        time_reverse_conj(0, 1) = 1.0;
        time_reverse_conj(1, 0) = -1.0;
        // the interaction and time reversal taken together:
        Eigen::Matrix2cd V_eff = time_reverse_conj
                                 * (m_material->polarizedSubtrSLD(wavevectors)
                                    - m_ambient_material->polarizedSubtrSLD(wavevectors));
        result = result * V_eff;
    }
    if (m_position)
        return result *= exp_I(m_position->dot(wavevectors.getQ()));
    return m_ff->thePolFF(wavevectors);
}
*/

double DecoratedFF::bottomZ(const IRotation* rotation) const
{
    RotMatrix transform = rotation ? rotation->getRotMatrix() : RotMatrix();
    /*
    if (m_rotMatrix) {
        std::cout << "DEBUG T " << transform << std::endl;
        std::cout << "DEBUG M " << *m_rotMatrix << std::endl;
        transform = transform * *m_rotMatrix;
    }
    */
    std::unique_ptr<const IRotation> total_rotation(IRotation::createRotation(transform));
    double result = m_ff->bottomZ(total_rotation.get());
    /*
    if (m_position)
        result += (rotation ? rotation->transformed(*m_position) : *m_position).z();
    */
    return result;
}

double DecoratedFF::topZ(const IRotation* rotation) const
{
    RotMatrix transform = rotation ? rotation->getRotMatrix() : RotMatrix();
    /*
    if (m_rotMatrix)
        transform = transform * *m_rotMatrix;
    */
    std::unique_ptr<const IRotation> total_rotation(IRotation::createRotation(transform));
    double result = m_ff->topZ(total_rotation.get());
    /*
    if (m_position)
        result += (rotation ? rotation->transformed(*m_position) : *m_position).z();
    */
    return result;
}
