//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Scattering/DecoratedFF.h
//! @brief     Defines interface class DecoratedFF.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef USER_API
#ifndef BORNAGAIN_SAMPLE_SCATTERING_DECORATEDFF_H
#define BORNAGAIN_SAMPLE_SCATTERING_DECORATEDFF_H

#include "Sample/Scattering/IDecoratableBorn.h"
#include <heinz/Vectors3D.h>

class Material;
class RotMatrix;

//! Encapsulates another formfactor and adds extra functionality
//! (a scalar factor, a position-dependent phase factor, ...).
//!
//! This class is designed according to the Decorator Pattern.
//! It inherits from IFormFactor _and_ has a member of type IFormFactor*.

class DecoratedFF : public IDecoratableBorn {
public:
    DecoratedFF(const IDecoratableBorn& ff);
    ~DecoratedFF() override;

    DecoratedFF* clone() const override;
    std::string className() const override { return "DecoratedFF"; }

    //! Sets the material of the scatterer
    void setMaterial(const Material& material);

    //! Sets the ambient material
    void setAmbientMaterial(const Material& ambient_material) override;

    void setPosition(const R3& position);
    void setRotMatrix(const RotMatrix& rotMatrix);

    complex_t theFF(const WavevectorInfo& wavevectors) const override;
#ifndef SWIG
    Eigen::Matrix2cd thePolFF(const WavevectorInfo& wavevectors) const override;
#endif

    double volume() const override { return m_ff->volume(); }

    double radialExtension() const override { return m_ff->radialExtension(); }

    double bottomZ(const IRotation* rotation) const override;

    double topZ(const IRotation* rotation) const override;

    const IDecoratableBorn* getFormFactor() const { return m_ff.get(); }

    complex_t formfactor(C3 q) const override { return m_ff->formfactor(q); }

protected:
    std::unique_ptr<const IDecoratableBorn> m_ff;

private:
    DecoratedFF(IDecoratableBorn* ff, const Material& material,
                const Material& ambient_material, const R3* position, const RotMatrix* rotMatrix);

    std::unique_ptr<const Material> m_material;
    std::unique_ptr<const Material> m_ambient_material;
    std::unique_ptr<const R3> m_position;
    std::unique_ptr<const RotMatrix> m_rotMatrix;
};

#endif // BORNAGAIN_SAMPLE_SCATTERING_DECORATEDFF_H
#endif // USER_API
