//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Scattering/IBornFF.cpp
//! @brief     Implements interface class IBornFF.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Scattering/IBornFF.h"
#include "Base/Util/Algorithms.h"
#include "Base/Vector/RotMatrix.h"
#include "Base/Vector/WavevectorInfo.h"
#include "Sample/Scattering/Rotations.h"
#include "Sample/Shapes/IShape3D.h"
#include <stdexcept>

namespace {

bool shapeIsContainedInLimits(const IFormFactor& formfactor, ZLimits limits,
                              const IRotation* rotation, R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return limits.zBottom() <= zbottom && ztop <= limits.zTop();
}

bool shapeOutsideLimits(const IFormFactor& formfactor, ZLimits limits, const IRotation* rotation,
                        R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return ztop <= limits.zBottom() || zbottom >= limits.zTop();
}

} // namespace


IBornFF::IBornFF() = default;

IBornFF::IBornFF(const NodeMeta& meta, const std::vector<double>& PValues)
    : IDecoratableBorn(meta, PValues)
{
}

IBornFF::~IBornFF() = default;

complex_t IBornFF::theFF(const WavevectorInfo& wavevectors) const
{
    return formfactor(wavevectors.getQ());
}

Eigen::Matrix2cd IBornFF::thePolFF(const WavevectorInfo& wavevectors) const
{
    return formfactor_pol(wavevectors.getQ());
}

double IBornFF::bottomZ(const IRotation* rotation) const
{
    if (!m_shape3D)
        throw std::runtime_error("Bug: Form factor has no m_shape3D, cannot compute bottom z");
    return BottomZ(m_shape3D->vertices(), rotation);
}

double IBornFF::topZ(const IRotation* rotation) const
{
    if (!m_shape3D)
        throw std::runtime_error("Bug: Form factor has no m_shape3D, cannot compute top z");
    return TopZ(m_shape3D->vertices(), rotation);
}

bool IBornFF::canSliceAnalytically(const IRotation* rotation) const
{
    return !rotation || rotation->zInvariant();
}

Eigen::Matrix2cd IBornFF::formfactor_pol(C3 q) const
{
    return formfactor(q) * Eigen::Matrix2cd::Identity();
}

IDecoratableBorn* IBornFF::createSlicedFormFactor(ZLimits limits, const IRotation* rot,
                                                  R3 translation) const
{
    if (shapeIsContainedInLimits(*this, limits, rot, translation))
        return createTransformedFormFactor(*this, rot, translation);
    if (shapeOutsideLimits(*this, limits, rot, translation))
        return nullptr;
    if (canSliceAnalytically(rot))
        return sliceFormFactor(limits, rot, translation);
    throw std::runtime_error("Slicing of " + className()
                             + " not supported for the given rotation!");
}

IDecoratableBorn* IBornFF::sliceFormFactor(ZLimits, const IRotation*, R3) const
{
    throw std::runtime_error(className() + "::sliceFormFactor error: not implemented!");
}

SlicingEffects IBornFF::computeSlicingEffects(ZLimits limits, const R3& position, double height)
{
    R3 new_position(position);
    double z_bottom = position.z();
    double z_top = position.z() + height;
    double dz_top = std::isinf(limits.zTop()) ? -1 : z_top - limits.zTop();
    double dz_bottom = std::isinf(limits.zBottom()) ? -1 : limits.zBottom() - z_bottom;
    ASSERT(dz_top >= 0 || dz_bottom >= 0);
    ASSERT(dz_bottom <= height);
    ASSERT(dz_top <= height);
    if (dz_bottom < 0)
        dz_bottom = 0;
    if (dz_top < 0)
        dz_top = 0;
    if (dz_bottom > 0)
        new_position.setZ(limits.zBottom());
    return {new_position, dz_bottom, dz_top};
}

double IBornFF::BottomZ(const std::vector<R3>& vertices, const IRotation* rotation)
{
    ASSERT(vertices.size());
    return BaseUtils::algo::min_value(
        vertices.begin(), vertices.end(), [&](const R3& vertex) -> double {
            return (rotation ? rotation->transformed(vertex) : vertex).z();
        });
}

double IBornFF::TopZ(const std::vector<R3>& vertices, const IRotation* rotation)
{
    ASSERT(vertices.size());
    return BaseUtils::algo::max_value(
        vertices.begin(), vertices.end(), [&](const R3& vertex) -> double {
            return (rotation ? rotation->transformed(vertex) : vertex).z();
        });
}
